package RCWS;

import java.util.LinkedList;
import java.util.List;

import models.Node;

public class EdgeHelper {
 /* AUXILIARY METHODS */
    
    /** Creates the edges in savings list according to the CWS heuristic 
     * @Precondition node 0 is the Depot
     * @Precondition if exists Warehouse in the route, it's unique and it's the node 1
     * @Postcondition If a savings edge has a Warehouse, it's the edge's origin */
    public static LinkedList<Edge> generateSavings(List<Node> nodesList) {
        LinkedList<Edge> edges = new LinkedList<Edge>();
        int nNodes = nodesList.size();
        Node depot = nodesList.get(0);
        for( int i = 1; i < nNodes - 1; i++ ) // node 0 is the depot
        {   
        	Node iNode = nodesList.get(i);
            for( int j = i + 1; j < nNodes; j++ )
            {   
            	Node jNode = nodesList.get(j);     
            	edges.add(createEdge(iNode,jNode,depot));
            }
        }
        return edges;
    }
    /** Creates an edge with costs and, optionally, the inverse edge
     * @param iNode
     * Origin
     * @param jNode
     * End
     * @param inv 
     * Should inverse be created?
     * @return The new created edge
     */
	public static Edge createEdge(Node iNode, Node jNode, boolean inv) {
		if (inv) return createEdge(iNode, jNode);
		
		Edge ijEdge = new Edge(iNode, jNode);
		setEdgeCostAndTime(ijEdge);
		
		return ijEdge;
	}
    /** Creates an edge and its inverse edge with costs and savings
     * relatives to a depot
     * @param iNode
     * @param jNode
     * @param depot
     * @return The new created edge
     */
    public static Edge createEdge(Node iNode, Node jNode, Node depot) { 
        
        Edge ijEdge = createEdge(iNode, jNode);
        ijEdge.setSavings(ijEdge.calcSavings(depot));
        Edge jiEdge = ijEdge.getInverseEdge();
        jiEdge.setSavings(jiEdge.calcSavings(depot));
        
        return ijEdge;
    }
    /** Creates an edge and its inverse edge with costs
     * @param iNode
     * @param jNode
     * @param depot
     * @return The new created edge
     */
    public static Edge createEdge(Node iNode, Node jNode) { 
    	// Creates ijEdge and assigns costs
	   Edge ijEdge = new Edge(iNode, jNode);
	   setEdgeCostAndTime(ijEdge);
	   // Creates jiEdge and assigns costs
	   Edge jiEdge = new Edge(jNode, iNode);
	   setEdgeCostAndTime(jiEdge);
	   // Set inverse edges
	   ijEdge.setInverse(jiEdge);
	   jiEdge.setInverse(ijEdge);
       
       return ijEdge;
    }
    /** Completes an edge with its inverse
     * 
     * @param ijEdge
     * @return The same edge with its inverse edge
     */
	public static Edge addInverseEdge(Edge ijEdge) {
		Edge jiEdge = new Edge(ijEdge.getEnd(), ijEdge.getOrigin());
		setEdgeCostAndTime(jiEdge);
		// Set inverse edges
		ijEdge.setInverse(jiEdge);
		jiEdge.setInverse(ijEdge);
		
		return ijEdge;
	}
    
    /** Creates edges connecting nodes with the depot,
        i.e., creates the edges (0,i) and (i,0) for all i > 0. */
    public static void generateDepotEdges(List<Node> nodesList) {   
    	Node depot = nodesList.get(0); // depot is always node 0
        int nNodes = nodesList.size();
        
        // Creates diEdge and idEdge, and set the corresponding costs
        for( int i = 1; i < nNodes; i++ ) // node 0 is depot
        {   
        	Node iNode = nodesList.get(i);
            // Creates edge depot-iNode, set costs, and assign it to iNode
        	Edge diEdge = createEdge(depot, iNode);
            iNode.setDiEdge(diEdge);
            iNode.setIdEdge(diEdge.getInverseEdge());
        }
    }
    /**
     * 
     */
    private static void setEdgeCostAndTime(Edge e) {
    	e.setLength(e.calcLength());
    	e.setTimeLength(e.calcTimeLength());
    }
}
