package RCWS;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import org.apache.commons.collections4.iterators.PermutationIterator;

import algorithm.Algorithm;
import algorithm.ArrayTools;
import main.RNG;
import map.SubMap;
import models.Availability;
import models.Node;
import models.SubCustomer;
import models.Depot;
import models.Test;
import models.VehType;
import models.Vehicle;
import models.Warehouse;
import data.Input;
import data.SubSol;

/**  
 * Biased-randomized version of the Clarke & Wright savings (CWS) heuristic.
 * @author Angel A. Juan - ajuanp(@)gmail.com
 * @version 130112
 */
public class RandCWS
{	 
    /** Generates a biased-randomized solution based on the CWS heuristic 
     * @param maxTime 
     * @param avails */
    public static SubSol solve(double savingsParameter, List<Node> nodes, Vehicle veh, List<Edge> savings, SubMap sm, float maxTime)
    {
    	int diffProducts = ((Depot)nodes.get(0)).getDiffProducts();
        /* 1. RESET VARIABLES */
        // dummySol resets isInterior and inRoute in nodes
        SubSol currentSol = generateDummySol(nodes, diffProducts, veh.getVehType(), maxTime);
        
        // Could be empty dummy solution because all customers deny to be delivered by vht
        if (!currentSol.getRoutes().isEmpty()) {
        	
	        Node depot = nodes.get(0);
	
	        Collections.sort(savings);
	        // Find the wh-route, if existing
	        WhRoute whRoute = null;
	        boolean found = false;
	     
	        for (int i=0; i < currentSol.getRoutes().size() && !found; i++) {
	        	Route r = currentSol.getRoutes().get(i);
	        	found =  r instanceof WhRoute;  
	        	if (found) whRoute = (WhRoute)r;
	        }
	        
	        /* 3. PERFORM THE EDGE-SELECTION & ROUTING-MERGING ITERATIVE PROCESS */
	        while( !savings.isEmpty()  )
	        {   
	            // 3.1. Select the next edge from the list (bias-randomized)
	            
	            int index = RNG.getNextIntGD(savings.size(), savingsParameter);    
	            Edge ijEdge = savings.get(index);
	            savings.remove(ijEdge); // remove edge from list
	            
	            // Force wh to be always in origin
	            if (ijEdge.getEnd().getInRoute() instanceof WhRoute) 
	            	ijEdge = ijEdge.getInverseEdge();
	            	
	            // 3.2. Determine the nodes i < j that define the edge
	            Node iNode = ijEdge.getOrigin();
	            Node jNode = ijEdge.getEnd();
	
	            // 3.3. Determine the routes associated with each node
	            Route iR = iNode.getInRoute();
	            Route jR = jNode.getInRoute();

	            if (iR != null && jR != null) { // Null routes ==> nodes doesn't accept veh type

		            // 3.4. If all necessary conditions are satisfied, merge
		            
		            // 3.4.1. Get an edge iE in iR containing nodes i and 0
		            Edge iE = getEdge(iR, iNode, depot); // iE is either (0,i) or (i,0)
		            
		            // 3.4.5. Get an edge jE in jR containing nodes j and 0
		            Edge jE = getEdge(jR, jNode, depot); // jE is either (0,j) or (j,0)
		            
		            double iRcosts = iR.getCosts();
		            double jRcosts = jR.getCosts();
		            
		            // Check merging conditions updating iR and jR routes if merge is ok
		            boolean ok =
							checkMergingConditions(iR, jR, ijEdge, iE, jE, sm, whRoute, veh, maxTime);

		            if( ok )
		            {                   
		            	currentSol.setCosts(currentSol.getCosts()-iRcosts-jRcosts);
		                currentSol.getRoutes().remove(jR);
		                currentSol.addCosts(iR);
		            }
	            }
	        }
	        /* 4. RETURN THE SOLUTION */
	        
	        int[] emptyLd = new int[veh.getNumCompartments()];
	        Arrays.fill(emptyLd, -1);
	        // Check if it can do the job for dummy routes
	        // If so, assign a possible load distribution; else, delete the route from the solution.
	        Iterator<Route> it = currentSol.getRoutes().iterator();
	        while (it.hasNext()) {
	        	Route r = it.next();
	       
	        	if(r.isDummy()) {
	        		int[] ld = emptyLd.clone();
	        		if (r instanceof WhRoute) 
	        			r.setLoadDistribution(ld);
	        		else {
		        		SubCustomer nodeDelivered = (SubCustomer)r.getEdges().getFirst().getEnd();
		        		
		        		int compartment = 0;
		        		int pending = nodeDelivered.getDemand();
		        		while (pending > 0 && compartment < veh.getNumCompartments()) {
		        			pending = pending - veh.getCapacity(compartment);      			
		        			ld[compartment] = nodeDelivered.getProduct();
		        			compartment++;
		        		}
		        		if (pending > 0) 
		        			it.remove(); // this vehicle type can't deliver this dummy route !!
		        		else 
		        			r.setLoadDistribution(ld);
	        		}
	        	}
	        }
        }
        return currentSol;
    }
    
    /** Constructs an initial dummy (usually not feasible) solution as described in the CWS
     *  heuristic: dummySol = { (0,i,0) / i in vrpNodesList }
     * During this process, inRoute and isInterior values are assigned 
     * This is only for nodes that accept vehicle type. That's why it can be not feasible
     * @param maxTime 
     */
    private static SubSol generateDummySol(List<Node> nodes, int diffProducts, VehType vht, float maxTime)
    { 
        SubSol sol = new SubSol(diffProducts);
        for( int i = 1; i < nodes.size(); i++ ) // i = 0 is the depot
        {   
        	Node iNode = nodes.get(i);
        	
        	boolean vhtAccepted = true;
        	if (iNode instanceof SubCustomer) {
        		SubCustomer sc = (SubCustomer)iNode;
        		vhtAccepted = sc.acceptsVehType(vht);
        		sc.clearInRoute();
        	}
        
        	if (vhtAccepted) {
        		
	            // Get diEdge and idEdge
	            Edge diEdge = iNode.getDiEdge();
	            Edge idEdge = iNode.getIdEdge();
	            // Create didRoute (and set corresponding total costs and demand)
	            Route didRoute;
	            if (iNode instanceof Warehouse) 
	            	didRoute = new WhRoute(diffProducts, vht);
	            else
	            	didRoute = new Route(diffProducts, vht);
	            
	            didRoute.getEdges().add(diEdge);
	            
	            
	            // update DEMAND
	            
	            if (diEdge.getEnd() instanceof SubCustomer) {
		            SubCustomer sc = (SubCustomer)diEdge.getEnd();
		            didRoute.setProductDemand(sc.getProduct(), didRoute.getProductDemand(sc.getProduct()) + sc.getDemand());
	            }
	            
	            // update COSTS
	            
	            didRoute.addCosts(diEdge);
	           
	            didRoute.setDummy(true);
	            
	            didRoute.getEdges().add(idEdge);
	            didRoute.addCosts(idEdge);
	            
	            if (!(didRoute.getTimeLength()+
	            		didRoute.calcProdTransferTime(Input.getFlow(), Input.getOpSetupTime())
	            		> maxTime)) {
		            // Update iNode properties (inRoute and isInterior)
		            iNode.setInRoute(didRoute); // save route to which node belongs
		            iNode.setIsInterior(false); // node is directly connected to depot
		            // Add didRoute to current solution
		            sol.getRoutes().add(didRoute); 
		             
		            sol.addCosts(didRoute);
		            
		            //VECTORIZED DEMAND
		            ArrayTools.inc(sol.getDemand(), didRoute.getDemand());
	//	            for(int prod = 0; prod < diffProducts; prod++) 
	//	            	sol.setProductDemand(prod,sol.getProductDemand(prod) + didRoute.getProductDemand(prod));
	            }
        	}
        }
        return sol;
    }

    /** Given aRoute, iNode and depot, returns the edge in aRoute which
     * contains iNode and depot (it will be the first or the last edge) */
    private static Edge getEdge(Route aRoute, Node iNode, Node depot)
    {   
        // Check if firstEdge in aRoute contains iNode and depot
        Edge firstEdge = aRoute.getEdges().get(0);
        Node origin = firstEdge.getOrigin();
        Node end = firstEdge.getEnd();
        if ( ( origin == iNode && end == depot )
                || ( origin == depot && end == iNode ) )
            return firstEdge;
        else
        {   int lastIndex = aRoute.getEdges().size() - 1;
            Edge lastEdge = aRoute.getEdges().get(lastIndex);
            return lastEdge;
        }
    }
    
    /** Checks merging conditions 
     * @param jE 
     * @param iE 
     * @param sm 
     * @param whRoute 
     * @param maxTime 
     * @param vht 
     * @Precondition jR is not a WhRoute */
    private static boolean checkMergingConditions(Route iR, Route jR, Edge ijEdge, Edge iE, Edge jE, SubMap sm, WhRoute whRoute, Vehicle veh, float maxTime)
    {
        // Condition 1: iR and jR are not the same route
        if( iR == jR )
            return false;

        // Condition 2: both nodes are exterior nodes in their respective routes
        Node iNode = ijEdge.getOrigin();
        Node jNode = ijEdge.getEnd();
        if( iNode.getIsInterior() == true || jNode.getIsInterior() == true )
            return false;
        
        // Condition 3: wh can be linked only if it's in a dummy routes 
        if (iNode instanceof Warehouse && !iR.isDummy())
        	return false;
       
        // Condition 4: It's possible to assign overall volumes of iR and jR demands to compartments of veh
    	int[] ld = calcLoadDistribution(veh.getVehType(), iR, jR);
        if (ld == null) 
        	return false;
        
        // Condition 5: Resulting route fits in time 
        //     Calculate resulting route 
        Route iR_ = iR.makeSimpleCopy();
        Route jR_ = jR.makeSimpleCopy();
        
    	iR_.getEdges().remove(iE);
    	jR_.getEdges().remove(jE);
    	    	
    	if( !(iR_.getEdges().get(0).getOrigin() instanceof Depot) ) 
            iR_.reverse();
    	
    	if( jR_.getEdges().get(0).getOrigin() instanceof Depot ) 
            jR_.reverse();
    	
    	iR_.getEdges().add(ijEdge);
    	
    	for( Edge e : jR_.getEdges() )
            iR_.getEdges().add(e);
    	
    	iR_.calcCosts();
    	//     Improve it
    	RouteCache.findImprovedNodeSeq(iR_);
   	
        //     route timeLength < maxTime - availability time of products
    	
    	iR_.setDemand(ArrayTools.sum(iR.getDemand(), jR.getDemand()));
    	iR_.setVehType(iR.getVehType());
    	float timeLength = iR_.getTimeLength() + iR_.calcProdTransferTime(Input.getFlow(), Input.getOpSetupTime());
        // Try with products already available
        float startTime = sm.calcStartTime(veh, iR_, ArrayTools.sum(iR.getDemand(), jR.getDemand()));
        if (startTime + timeLength > maxTime) 
        	// Last chance is wh-route in current subsolution
        	if (whRoute == null || whRoute == iR_)
        		return false;
        	else {
	        	startTime = sm.calcStartTime(veh, whRoute);
	        	if (startTime + whRoute.getTimeLength() + timeLength > maxTime)
	        		return false;
        	}
        
        // check is OK, merge both routes into iR 
                
        iR.setLoadDistribution(ld);
        
        jR.removeCosts(jE);
        
        jNode.setInRoute(iR);
        
        for (Edge e : jR_.getEdges()) { 
       
            e.getEnd().setInRoute(iR);
    	}
        
        ArrayTools.inc(iR.getDemand(), jR.getDemand());
        
        iR.setDummy(false);
                
        iR.setEdges(iR_.getEdges());
        
        iR.calcCosts();
        
        // Update nodes according new position in route
        
        int nEdges = iR.getEdges().size();
        iR.getEdges().get(1).getOrigin().setIsInterior(false);
        iR.getEdges().get(nEdges-1).getOrigin().setIsInterior(false);
        for (int i=2; i<nEdges-1; i++)
        	iR.getEdges().get(i).getOrigin().setIsInterior(true);
        
		return true;
    }  
    /**
     * Calculates a possible load distribution on compartments of a given vehicle type
     * to acommodate total demand of two routes
     * @param vht The vehicle type
     * @param r1 One route
     * @param r2 The other route
     * @return load distribution of products in compartments, or null if it's not possible
     */
    private static int[] calcLoadDistribution(VehType vht, Route r1, Route r2) {
    	
    	class Compartment { private int no; private int capacity; }
    	
    	List<Compartment> compList = new LinkedList<Compartment>();
    	int[] prodInComp = new int[vht.getNumCompartments()];
    	
        for (int i=0; i<vht.getNumCompartments(); i++) {
        	Compartment c = new Compartment();
        	c.capacity = vht.getCapacity(i);
        	c.no = i;
        	compList.add(c);
        	
        	prodInComp[i] = -1; // No product in compartments, initially
        }
        
        PermutationIterator<Compartment> pit = new PermutationIterator<Compartment>(compList);
        int demand;
        do {
        	List<Compartment> permutedCaps = pit.next();
        	int prod = 0;
        	int comp = 0;
        	
        	do {
        		demand = r1.getProductDemand(prod) + r2.getProductDemand(prod);
        		
        		while (demand > 0 && comp < vht.getNumCompartments()) {
        			demand = demand - permutedCaps.get(comp).capacity;
        			prodInComp[permutedCaps.get(comp).no] = prod;
        			comp++;
        		}
        		prod++;
        		
        	} while (demand <= 0 && prod < r1.getDemand().length); // Product can be loaded and
        	                                              // There are more products to load
        } while (pit.hasNext() && demand > 0); // Try with next permutation if available and 
        									   // still pending demand to be placed
        
        if (demand > 0) return null;
        
        return prodInComp;
    }
}