package RCWS;

import java.util.Arrays;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;

import map.MapTools;
import models.Delivery;
import models.Depot;
import models.Node;
import models.VehType;

/**
 * @author Angel A. Juan - http://ajuanp.wordpress.com
 * @version 130119
 */
public class Route {

	/* INSTANCE FIELDS & CONSTRUCTOR */
	private static long nInstances = 0; // number of instances
	private long id = 0; // route id
	private LinkedList<Edge> edges = new LinkedList<Edge>(); // evolving
	private double costs; // route total costs
	private double length; // route total length
	private int[] demand; // route demand
	private VehType vht;

	private boolean dummy = false;
	private int[] loadDistribution;  // product loaded in each vehicle compartment
	private List<Delivery> deliveries = new LinkedList<Delivery>(); // deliveries to each node 
	private float startTime;
	private float timeLength;
	
	public Route(int workingProducts, VehType vht) {
		nInstances++;
		this.id = nInstances;

		this.vht = vht;
		this.demand = new int[workingProducts];	
		this.costs = vht.getFixedCost();
		this.length = 0.0;
		this.timeLength = 0f;
	}
	/**
	 * Constructor to create a simplified Route copy for caching and merge simulation pourposes
	 * @param aRoute
	 */
	public Route(Route aRoute) {
		nInstances++;
		this.id = nInstances;
		this.edges = new LinkedList<Edge>(aRoute.getEdges());
		this.length = aRoute.getLength();
		this.timeLength = aRoute.getTimeLength();
	}

	public void setProductDemand(int product, int d) {
		demand[product] = d;
	}
	
	public void setDemand(int[] demand) {
		this.demand = demand;
	}

	public VehType getVehType() {
		return vht;
	}

	public void setVehType(VehType vht) {
		this.vht = vht;
	}

	public double getCosts() {
		return costs;
	}
	
	public double getLength() {
		return length;
	}

	public int[] getDemand() {
		return demand;
	}

	public int getProductDemand(int product) {
		return demand[product];
	}

	public LinkedList<Edge> getEdges() {
		return edges;
	}
	
	public void setDummy(boolean b) {
		dummy = b;
	}

	public boolean isDummy() {
		return dummy;
	}


	public long getId() {
		return id;
	}
	
	public void setLoadDistribution(int[] loadDistribution) {
		this.loadDistribution = loadDistribution;
		
	}

	public int[] getLoadDistribution() {
		return loadDistribution;
	}

	public List<Delivery> getDeliveries() {
		return deliveries;
	}

	public void setDeliveries(List<Delivery> deliveries) {
		this.deliveries = deliveries;
	}

	/* AUXILIARY METHODS */

	/** Reverses a route, e.g. (0 -> 2 -> 6 -> 0) becomes (0 -> 6 -> 2 -> 0) */
	public void reverse() {
		for (int i = 0; i < edges.size(); i++) {
			Edge e = edges.get(i);
			Edge invE = e.getInverseEdge();
			edges.remove(e);
			edges.add(0, invE);
		}
	}

	public void addEdge(Edge anEdge) {
		if (edges.contains(anEdge) == false)
			edges.add(anEdge);
	}

	public void removeEdge(Edge anEdge) {
		if (edges.contains(anEdge) == true)
			edges.remove(anEdge);
	}

	public void removeCosts(Edge anEdge) {
		if (vht != null)
			costs -= anEdge.getLength() * vht.getVaribleCost();
		length -= anEdge.getLength();
		timeLength -= anEdge.getTimeLength();
	}

	public void addCosts(Edge anEdge) {
		if (vht != null)
			costs += anEdge.getLength() * vht.getVaribleCost();
		length += anEdge.getLength();
		timeLength += anEdge.getTimeLength();
	}
	
	public void calcCosts() {
		length = 0.;
		if (vht != null)
			costs = vht.getFixedCost();
		timeLength = 0f;
		
		for (Edge e : edges) 
			addCosts(e);
	}
	
	public void setEdges(LinkedList<Edge> edges) {		
		this.edges = edges;
	}

	// COMPARATORS
	public static final Comparator<Route> LESS_COST_ROUTES = new Comparator<Route>() {

		@Override
		public int compare(Route o1, Route o2) {
			if (o1.getCosts() < o2.getCosts() ) return -1;
			if (o1.getCosts() > o2.getCosts() ) return 1;
			return 0;
		}
	};
	
	public static final Comparator<Route> LESS_COST_PER_EDGE_ROUTES = new Comparator<Route>() {
		@Override
		public int compare(Route o1, Route o2) {
			double ratio1 = o1.getCosts() / o1.getEdges().size();
			double ratio2 = o2.getCosts() / o2.getEdges().size();
			if (ratio1 < ratio2) return -1;
			if (ratio1 > ratio2) return 1;
			return 0;
		}
	};
	public static final Comparator<Route> MORE_EDGES_ROUTES = new Comparator<Route>() {
		@Override
		public int compare(Route o1, Route o2) {
			
			return o2.getEdges().size() - o1.getEdges().size();
		}
	};

	public static final Comparator<Route> WHROUTES_FIRST = new Comparator<Route>() {
		@Override
		public int compare(Route o1, Route o2) {
			boolean b1 = o1 instanceof WhRoute;
			boolean b2 = o2 instanceof WhRoute;
			if (b2 == b1) return 0;
			if (b2) return 1;
			return -1;
		}
	};
	
	public List<Node> getNodes() {
		List<Node> tr = new LinkedList<Node>();
		for(Edge e: edges){
			if(!tr.contains(e.getOrigin()))tr.add(e.getOrigin());
			if(!tr.contains(e.getEnd()))tr.add(e.getEnd());
		}
		return tr;
	}
	
	public float getStartTime() {
		return startTime;
	}

	public void setStartTime(float f) {
		this.startTime = f;
	}

	public float getTimeLength() {
		return timeLength;
	}	
		
	public void addProdTransferTime(float flow, float opSetupTime) {
		
		timeLength += calcProdTransferTime(flow, opSetupTime);
	}
	
	public float calcProdTransferTime(float flow, float opSetupTime) {
		
		int totalDemand = 0;
		
		for(int demand : this.demand) 
			totalDemand += demand;
		
		return  2 * totalDemand / flow + edges.size() * opSetupTime;
	}
	
	public Route copyEdgesFrom(Route r) {
		this.edges = new LinkedList<Edge>(r.getEdges());
		this.length = r.getLength();
		if (vht != null)
			this.costs = length * vht.getVaribleCost() + vht.getFixedCost();
		this.timeLength = r.getTimeLength();
		return this;
	}
	
	public Route makeSimpleCopy() {	
		return new Route(this);
	}	
	
	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("\n\t\t\t\t\t\t\tRoute [")
				.append(this.toString2())
				.append("\n\t\t\t\t\t\t\t]");
		return builder.toString();
	}
	
	public String toString2() {
		StringBuilder builder = new StringBuilder();
		builder.append("id=").append(id).append(",")
				.append("\n\t\t\t\t\t\t\t\tedges=[")
				.append("\n\t\t\t\t\t\t\t\t\t").append(MapTools.str(edges, 9, 100))
				.append("\n\t\t\t\t\t\t\t\t]")
				.append(",\n\t\t\t\t\t\t\t\tcosts=").append(String.format(Locale.US, "%.2f",costs))
				.append(", length=").append(String.format(Locale.US, "%.2f",length))
				.append(", demand=").append(Arrays.toString(demand))
				.append(",\n\t\t\t\t\t\t\t\tloadDistribution=")
				.append(Arrays.toString(loadDistribution))
				.append(",\n\t\t\t\t\t\t\t\tdeliveries=[");
				if (!deliveries.isEmpty())
					builder.append("\n\t\t\t\t\t\t\t\t\t")
					.append(MapTools.str(deliveries, 9, 100))
					.append("\n\t\t\t\t\t\t\t\t");
				builder.append("]")
				.append(",\n\t\t\t\t\t\t\t\tstartTime=").append(String.format(Locale.US, "%.2f",startTime))
				.append(", timeLength=").append(String.format(Locale.US, "%.2f",timeLength))
				.append(", arrivalTime=").append(String.format(Locale.US, "%.2f",startTime+timeLength));
		return builder.toString();
	}


	/**
	 * For debugging pourposes. 
	 * @return The route is not correct
	 */
	public boolean check() {
		boolean err = false;
		boolean first = true;
		Node nextOrigin = null;
		for (Edge e: this.getEdges()){
			
			if (first) {
				
				if (!(e.getOrigin() instanceof Depot)) 
					err = true;
				
				first = false;
			} else 
				if (e.getOrigin() != nextOrigin)  
					err = true;
				
			if (err) break;
			nextOrigin = e.getEnd();
		}
		return (err || !(nextOrigin instanceof Depot)) ;
	
	}
}