package RCWS; 

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import models.Depot;
import models.Node;
import models.Warehouse;
import data.SubSol;



/***********************************************************************************
 * Project SimORouting - RouteCache.java
 * 
 * Two routes are said to be "nodes-equivalent" if they contain the same set of 
 *  nodes (notice that a set of nodes can be sorted in different ways).
 *  
 * This class uses a hash table to save the best known-so-far route for each set of
 *  nodes. Therefore, given a route we can try to improve it by looking up at the
 *  hash table, which could contain a better nodes-equivalent route.
 * 
 * Timestamp (YYMMDD_HHMM): 090202
 * David Masip & Angel Juan - dmasipr(@)uoc.edu & ajuanp(@)gmail.com
 **********************************************************************************/

public class RouteCache0 
{
	
	/******************************************************************************* 
     * INSTANCE FIELDS 
     * ****************************************************************************/
	
	// Declare and define some basic parameters of the hash table
	private static Hashtable<Integer, Route> tableR;
	private static Hashtable<Integer, SubSol> tableS;
	private static int initialCapacity = 1000; 
    private static float loadFactor = (float) 0.8;
    private static int tableSize;
    /******************************************************************************* 
     * CLASS CONSTRUCTOR 
     ******************************************************************************/
	
	public RouteCache0()
	{
		// Initialize the hash table
		tableR = new Hashtable<Integer, Route>(initialCapacity, loadFactor);
		tableS = new Hashtable<Integer, SubSol>(initialCapacity, loadFactor);
		tableSize = 0;
	}
	

	

	/******************************************************************************* 
     * PUBLIC METHOD getTableRSize()
     ******************************************************************************/
	
	public int getTableRSize()
	{
		tableSize = tableR.size();
		return tableSize;
	}
	
	
	/******************************************************************************* 
     * PUBLIC METHOD getTableSSize()
     ******************************************************************************/
	
	public int getTableSSize()
	{
		tableSize = tableS.size();
		return tableSize;
	}
	
	
	/******************************************************************************* 
     * PRIVATE METHOD findImprovedRoute()
     * Given aRoute it returns either an improved route (using the same nodes) from 
     *  the hash table, or simply aRoute if there isn't a better route in the table
     * ****************************************************************************/
	
	public static void findImprovedNodeSeq(Route aRoute)
	{
		
		// 1. Obtain the set of nodes contained in aRoute
		int nodes = getNodesFromRoute(aRoute);
		
		// 2. Obtain, if exists, the cached route containing the same set of nodes
		Route cachedRoute = (Route) tableR.get(nodes);	
		
		// 3. If it still does not exist, improve aRoute, save a copy of it in cache and return it
		if (cachedRoute == null) {				
			improveNodesOrder(aRoute);
			tableR.put(nodes, new Route(aRoute));
		}
		
		// 4. If it already exist, update the cache if convenient
		else {
		
			// 4.1. If aRoute is better than cachedRoute, improve aRoute, 
			//      save a copy of it
			if ( aRoute.getLength() < cachedRoute.getLength() )
			{
				tableR.remove(nodes);
				improveNodesOrder(aRoute);
				tableR.put(nodes, new Route(aRoute));		
			}
			// 4.2. If cachedRoute is better than aRoute, copy 
			// 		cachedRoute's edges, costs, length an timeLength
			else
				aRoute.copyEdgesFrom(cachedRoute);
		}
	}

		
	/******************************************************************************* 
     * PRIVATE METHOD getNodesFromRoute()
     * Given a route (e.g.: 0-1-2-23-15-0), this method returns a string containing
     *  its nodes sorted and separated by semicolons (e.g. "0:0:1:2:15:23"). This
     *  string can then be used as a key in a hash table.
     * ****************************************************************************/
	
	private static int getNodesFromRoute(Route aRoute)
	{
		// 1. Construct a list edges containing all edges in aRoute
		List<Edge> edges = aRoute.getEdges();
		
		// 2. Construct an array to save all nodesID contained in aRoute
		//	  (e.g.: route 0-1-2-23-15-0 has 5 edges and 6 nodes)
		int nodesID[] = new int[edges.size() + 1];
		
		// 3. Select the first edge in the list of edges
		Edge anEdge = edges.get(0);
		
		// 4. Get the ID of the first node in anEdge (depot)
		nodesID[0] = anEdge.getOrigin().getCachingId(); 
		
		// 5. Look for every node in aRoute and save its corresponding ID
		int i = 0;
		for( i = 0; i < aRoute.getEdges().size() - 1; i++ )
		{
			nodesID[ i + 1 ] = anEdge.getEnd().getCachingId();
			anEdge = edges.get( i + 1 );
		}
		
		// 6. Close the array by adding node 0 as the last node (again depot)
		nodesID[ i + 1 ] = anEdge.getEnd().getId();
		
		// 7. Sort nodes by ID (e.g.: (0,0,1,2,15,23)
		Arrays.sort(nodesID);
		
		// 8. Construct a string by adding each node in nodesID separated by ":"
		//    (e.g.: ":0:0:1:2:15:23")
		String nodes = "";
		for( i = 0; i < nodesID.length; i++ )
			nodes = nodes + ":" + nodesID[i];
		
		// 9. Return the hash code associated with the resulting string
		int hashCode = nodes.hashCode();
		return hashCode;
	}
	
	

	/******************************************************************************* 
     * PRIVATE METHOD improveNodesOrder()
     * Given aRoute, this method tries to sort its nodes in a more efficient way. 
     *  (2-opt algorithm)
     * ****************************************************************************/

	public static void improveNodesOrder(Route aRoute)
	{
		List<Edge> edges = aRoute.getEdges();
		int nEdges = edges.size();
		if (nEdges > 3) {
			boolean toImprove = false;
			boolean improved = false;
	
			do {
				toImprove = false;
				Edge bestE1 = null;
				Edge bestE3 = null;
				Edge bestE1b = null;
				Edge bestE3b = null;
				double bestImpro = 0.;
				
				for (int i = 0; i < nEdges; i++)
				{			
					Edge e1 = edges.get(i);
					Node originE1 = e1.getOrigin();
					Node endE1 = e1.getEnd();
					// Check depot -> wh or wh -> depot are not involved in modifiable edges
					if (originE1 instanceof Depot || originE1 instanceof Warehouse) 
						if (endE1 instanceof Depot || endE1 instanceof Warehouse) 
							continue;
					
					// Second edge: e_(i+2) ... e_(i-2) (mod nEdges)
					for (int j = (i+2) % nEdges; j != (i-1+nEdges) % nEdges; j = (j+1) % nEdges)
					{
						// Get the current way. 
						Edge e3 = edges.get(j);
						Node originE3 = e3.getOrigin();
						Node endE3 = e3.getEnd();
						if (originE3 instanceof Depot || originE3 instanceof Warehouse) 
							if (endE3 instanceof Depot || endE3 instanceof Warehouse) 
								continue;
						
						double currentLength = e1.getLength() + e3.getLength();
						
						// Construct the alternative way				
						Edge e1b = EdgeHelper.createEdge(originE1, originE3, false);					
						Edge e3b = EdgeHelper.createEdge(endE1, endE3, false);
						
						double alterLength = e1b.getLength() + e3b.getLength();
						
						// Compare both ways 
						if ( alterLength < currentLength )
						{
							double impro = currentLength - alterLength;
							if (impro > bestImpro) {
								bestImpro = impro;
								bestE1 = e1;
								bestE3 = e3;
								bestE1b = EdgeHelper.addInverseEdge(e1b);
								bestE3b = EdgeHelper.addInverseEdge(e3b);
								toImprove = true;
							}
						}
					}
				} 
				
	//			Improve
				if (toImprove) {
					int i = edges.indexOf(bestE1);
					int j = edges.indexOf(bestE3);
					
					edges.set(i, bestE1b);
					aRoute.removeCosts(bestE1);
					aRoute.addCosts(bestE1b);
					
					edges.set(j, bestE3b);
					aRoute.removeCosts(bestE3);
					aRoute.addCosts(bestE3b);
					
					// Move intermediate edges to an auxiliary list (inverted)
					List<Edge> aux = new LinkedList<Edge>();
					int nMoves = (j+nEdges - i) % nEdges - 1; // Number of edges to move
					int pos = (i+1) % nEdges;
					for (int k = 0; k < nMoves; k++) {
						Edge e = edges.remove(pos);
						if (pos == edges.size()) pos = 0;
						Edge eb = e.getInverseEdge();
						aux.add(0, eb);
					}
					
					// Return the edges from the auxiliary list, after bestE1b
					pos = (edges.indexOf(bestE1b) + 1) % nEdges;
					edges.addAll(pos, aux);
					improved = true;
				}
			} while (toImprove);
			
			if (improved) {
				// Reorganize the route to start and end in a depot
				LinkedList<Edge> edges2 = new LinkedList<Edge>();
				Iterator<Edge> it = edges.iterator();
				Edge edge = it.next();
				Node origin = edge.getOrigin();
				while (!(origin instanceof Depot)) {
					it.remove();
					edges2.add(edge);
					edge = it.next();
					origin = edge.getOrigin();
				}
				edges.addAll(edges2);
				
				// Invert route if last edge is WH -> Depot	
				if (edges.get(edges.size()-1).getOrigin() instanceof Warehouse) {
					
					edges2 = new LinkedList<Edge>();
					for (Edge e : edges) 
						edges2.add(0, e.getInverseEdge());
					
					aRoute.setEdges(edges2);
				}
			}
		}
		aRoute.calcCosts();
	}
	
	/******************************************************************************* 
     * PRIVATE METHOD findImprovedSol()
     * Given aSol it returns either an improved sol (using the same nodes) from 
     *  the hash table, or simply aSol if there isn't a better sol in the table
     * ****************************************************************************/
	
	public static SubSol findImprovedSol(SubSol aSol)
	{
		// 1. Obtain the set of nodes contained in aSol
		int nodes = getNodesFromSol(aSol);
		
		// 2. Obtain, if exists, the cached sol containing the same set of nodes
		SubSol cachedSol = (SubSol) tableS.get(nodes);	
		
		// 3. If it does not exist already, save aSol in cache and return it
		if (cachedSol == null)
		{
			tableS.put(nodes, aSol);
			return aSol;
		}
		
		// 4. If it exist already, update the cache if convenient
		else
		{
			// 4.1. If aSol is better than cachedSol, save aSol and return it
			if ( aSol.getCosts() < cachedSol.getCosts() )
			{
				tableS.remove(nodes);
				tableS.put(nodes, aSol);		
				return aSol;
			}
			// 4.2. If cachedSol is better than aSol, return cachedSol
			else
				return cachedSol;
		}
	}
	/******************************************************************************* 
     * PRIVATE METHOD getNodesFromSol()
     * Given a subsol (e.g.: 0-1-23-0-15-12-0), this method returns a string 
     *  containing its nodes sorted and separated by semicolons (e.g. 
     *  "0:0:0:1:12:15:23"). This string can then be used as a key in a hash table.
     * ****************************************************************************/
	
	private static int getNodesFromSol(SubSol aSol)
	{
		// 1. Construct a list edges containing all edges in aSol
		ArrayList<Edge> edges = new ArrayList<Edge>();
		for ( int i = 0; i < aSol.getRoutes().size(); i++ )
		{
			Route iRoute = aSol.getRoutes().get(i);
			edges.addAll(iRoute.getEdges());
		}
			
		// 2. Construct an array to save all nodesID contained in aSol
		//	  (e.g.: sol 0-1-23-0-15-12-0 has 6 edges and 7 nodes)
		int nodesID[] = new int[edges.size() + 1];
		
		// 3. Select the first edge in the list of edges
		Edge anEdge = edges.get(0);
		
		// 4. Get the ID of the first node in anEdge
		nodesID[0] = anEdge.getOrigin().getId(); 
		
		// 5. Look for every node in aSol and save its corresponding ID
		int i = 0;
		for( i = 0; i < edges.size() - 1; i++ )
		{
			nodesID[ i + 1 ] = anEdge.getEnd().getId();
			anEdge = edges.get( i + 1 );
		}
		
		// 6. Close the array by adding node 0 as the last node
		nodesID[ i + 1 ] = anEdge.getEnd().getId();
		
		// 7. Sort nodes by ID (e.g.: (0,0,0,1,12,15,23)
		Arrays.sort(nodesID);
		
		// 8. Construct a string by adding each node in nodesID separated by ":"
		//    (e.g.: ":0:0:0:1:12:15:23")
		String nodes = "";
		for( i = 0; i < nodesID.length; i++ )
			nodes = nodes + ":" + nodesID[i];
	
		// 9. Return the hash code associated with the resulting string
		int hashCode = nodes.hashCode();
		return hashCode;
	}
}	


