package algorithm;

import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Random;
import java.util.Set;

import main.RNG;
import map.CompleteMap;
import map.MapTools;
import map.Savings;
import map.SubMap;
import models.Delivery;
import models.Depot;
import models.Customer;
import models.Node;
import models.SubCustomer;
import models.Test;
import models.Vehicle;
import models.Warehouse;
import RCWS.Edge;
import RCWS.EdgeHelper;
import RCWS.RandCWS;
import RCWS.Route;
import RCWS.RouteCache;
import RCWS.WhRoute;
import data.Input;
import data.Solution;
import data.SubSol;

public class Algorithm {
	private final Test test;
	private final Input input;

	public Algorithm(Test test, Input input) {
		super();
		this.test = test;
		this.input = input;
	}

	public Test getTest() {
		return test;
	}

	public Input getInput() {
		return input;
	}

	public Solution solve() {
		// 1. PRELIMINARY
		// 1.1 Savings
		Savings savings = MapTools.calcDepotSavings(input);
		// 1.2 Global list of customers													
		List<Customer> nodes = input.getCustomers(); 

		// 1.3 Global list of depots
		List<Depot> depots = input.getDepots();   
		// 1.4 subCustomers for each customer
		for (Customer c : nodes) 
			for (int i = 0; i < c.getProductDemands().length; i++) 
				if(c.getDemand(i)>0) 
					c.setSubCustomer(new SubCustomer(c, i));
		
		// 2. Multi Start Process
		long totalTime = 1000000000l * test.getTime(); // in nanoseconds
		long startProcTime = System.nanoTime();
		
		Solution bestSol = null;
		int nSolutionsFound = 0;
	
		//while (nSolutionsFound < 100000){
		while (System.nanoTime() - startProcTime < totalTime) {
			
			Solution newSol = new Solution();

			// 2.1 Generate map assignment

			CompleteMap map = genRandAssignmentMap(nodes, depots, savings);

			if (map != null){
				
				// 2.2 Solve each map
				boolean feasibility = true;

				Iterator<SubMap> it = map.getSubMaps().iterator(); 
				while (it.hasNext() && feasibility) {
					SubMap sm = it.next();

					// solve each submap
					SubSol partialSol = biasedRandRichCws(sm, input);

					feasibility = isFeasible(partialSol, sm);
					
					if (feasibility) {
						
						// 2.2.4 Improve it
						SubSol improvedSol = RouteCache.findImprovedSol(partialSol);

						// 2.2.3 Add to Solution
						newSol.addSubSol(improvedSol);
					}
				}
				// 2.3 Compare with best...
				
				if (feasibility) {
					nSolutionsFound += 1;

					if (bestSol == null || newSol.getCost() < bestSol.getCost()) {
						bestSol = newSol;
						System.out.println(
							String.format(Locale.US, 
								"New best found. Cost: %.2f, time length: %.2f among %d solutions in %.2f s", 
								bestSol.getCost(), bestSol.getTimeLength(),	nSolutionsFound,
								(float)(System.nanoTime() - startProcTime)/1000000000L));
					}
				}
			}
		}
		System.out.println("Found " + nSolutionsFound + " solutions in " +
				(float)(System.nanoTime() - startProcTime)/1000000000L + " s");
		return bestSol;
	}
	/**
	 * 
	 * @param sol
	 * @param sm
	 * @return
	 */
	private boolean isFeasible(SubSol sol, SubMap sm) {
				
		int numDeliveries = 0;
		for (Route r : sol.getRoutes())
			numDeliveries += r.getDeliveries().size();
		
		return numDeliveries == sm.getDeliveryNeeds().size() &&
				sm.enoughInventory(sm.getDepot().getProductsReserve());
	}

	/**
	 * Creates a subsolution composed of routes to deliver a submap. First converts 
	 * each customer into several subcustomers (one for each product), then calculates the 
	 * routes by CWS method and finally rejoins subcustomers into original customers for each
	 * route
	 * 
	 * @param sm The submap
	 * @param input The problem's input data
	 * @return The subsolution, or null if it cannot be found
	 */
	private SubSol biasedRandRichCws(SubMap sm, Input input) {

		// Get depot, wh and depot vehicles
		Depot depot = sm.getDepot();
		
		List<Vehicle> vehicles = depot.getVehicles();
		
		Random rng = RNG.getrng();
		
		// Create a list with one subcustomer for each customer's demand of product
		
		List<SubCustomer> subcustomers = customers2subCustomers(sm);
		
		List<Route> goodRoutes = new LinkedList<Route>(); 
		
		//Collections.sort(vehicles, Vehicle.LESS_TOTAL_CAPACITY); // Sort vehicles by total capacity
		Collections.shuffle(vehicles, rng);

		int[] pendDemand = sm.getDemand().clone();
		//ArrayTools.inc(pendDemand, depot.getProductsReserve());
		
		List<Vehicle> eligibleVeh = new LinkedList<Vehicle>(vehicles); 
		
		boolean end = false;

		while (( !subcustomers.isEmpty() || !sm.enoughInventory(ArrayTools.sum(pendDemand, depot.getProductsReserve())) )&& 
				 !end && !eligibleVeh.isEmpty()) {
			// Get a random vehicle among eligibles

			Vehicle veh = eligibleVeh.get(RNG.getNextIntUD(eligibleVeh.size()));
			// with this vehicle type, create set of routes trying to deliver all subcustomers 
			List<Node> nodes = new LinkedList<Node>(subcustomers);  
			nodes.add(0, depot);
			
			// Add a wh if needed: 			
			if (!sm.enoughInventory(pendDemand) ||     // when not enough inventory to deliver pending customers
					subcustomers.isEmpty() ) { // when demand is covered so, if we are here it's because minimum inventory is not achieved
				Warehouse wh = sm.getAwh(nodes, input.getWarehouses());
				end = true;
				if (wh != null) {
					nodes.add(1, wh);
					// Remove vehicles from eligibles until a wh-capable one is found
					while (!veh.isWhCapable() && !eligibleVeh.isEmpty()) {
						eligibleVeh.remove(veh);
						if (!eligibleVeh.isEmpty()) 
							veh = eligibleVeh.get(RNG.getNextIntUD(eligibleVeh.size()));
					}
					end = !veh.isWhCapable(); 
				}
			}
			
			if (!end) {
				
				// At this point, Depot has index 0 in nodes and Warehouse, if it exists, has index 1
				// This is a precondition for generateSavings()
				
				// Calculate the savings
				List<Edge> savings = EdgeHelper.generateSavings(nodes);
				EdgeHelper.generateDepotEdges(nodes);
				// Compute homogeneous fleet solution 
				SubSol hfSol = RandCWS.solve(test.getSavingsParameter(), nodes, veh, savings, sm, input.getMaxTime());
			
				// Process generated routes
				List<Route> currentRoutes = hfSol.getRoutes();
				
				// Randomize routes, but keeping wh-routes at top
				Collections.shuffle(currentRoutes, rng);
				Collections.sort(currentRoutes, Route.WHROUTES_FIRST);

				boolean addedRoutes = false; // to check if good routes have been generated with veh
				
				for (Route r : currentRoutes) {

					// Calculate "optimum" purchase in wh
					if (r instanceof WhRoute) 
						((WhRoute)r).calcPurchase(pendDemand, sm);
					// Add times due to load/unload and operation setup
					r.addProdTransferTime(Input.getFlow(), Input.getOpSetupTime());
					// set route start time according current needs
					boolean ok = sm.putRoute(r, veh); 
					
					if (ok) {	
						// Remove all subcustomers supplied by this route
						// and clear their demands from pending demands
						for (Edge e : r.getEdges()) 
							if (e.getEnd() instanceof SubCustomer) {
								SubCustomer sc = (SubCustomer)e.getEnd();
								subcustomers.remove(sc);
								pendDemand[sc.getProduct()] -= sc.getDemand();
							}
								
						// Convert the SubCustomers route to a Customers one
						r = subcustomers2customers(r);
						
						// Store the route
						goodRoutes.add(r);
						addedRoutes = true;
						if (r instanceof WhRoute) {
							WhRoute whR = (WhRoute) r;
							// Restore vehicles to try with, as new Availability is got
							eligibleVeh = new LinkedList<Vehicle>(vehicles);
							// New Availability, as vehicle will supply the remaining goods to depot
							sm.addAvail(whR);
						}
					} 
				}
				if (!addedRoutes) {
					// Remove vehicle from eligible ones
					eligibleVeh.remove(veh);
				}
			}
		}
		
		return constructSubSolution(goodRoutes, sm);
	}
	
	/**
	 * From the submap's list of nodes, creates a new list changing each customer by several
	 * subcustomers, one for each product demand.
	 *   
	 * @param sm The Submap
	 * @return The created list
	 */
	private List<SubCustomer> customers2subCustomers(SubMap sm) {
		List<SubCustomer> subNodes = new LinkedList<SubCustomer>();
		for (Node n : sm.getAssignedNodes()) {
			Customer c = (Customer)n;
			for (int i=0; i< c.getProductDemands().length; i++) 
				if(c.getDemand(i) > 0) 
					subNodes.add(c.getSubCustomer(i));
		}
		return subNodes;
	}

	/**
	 * Converts a subcustomers route into a customers one, removing possible loops.
	 * At same time, it creates the route's delivery list
	 * @param r The route
	 * @return The converted route
	 */
	private Route subcustomers2customers(Route r) {


		// Nothing to do with dummy wh-routes
		if (r instanceof WhRoute && r.isDummy()) return r;
		
		Set<Customer> visitedNodes = new HashSet<Customer>();

		Iterator<Edge> it = r.getEdges().iterator();
		Node depot = r.getEdges().get(0).getOrigin();
		Node nextOrigin = depot;
		
		
		// if it's a wh-route, skip wh
		if (r instanceof WhRoute) {
			nextOrigin = r.getEdges().get(1).getOrigin();
			it.next();
		}

		while (it.hasNext()) {
			Edge e = it.next();
			
			Node jN = e.getEnd();
			
			if (jN instanceof Depot) {

				Customer iNc = (Customer) nextOrigin;
				iNc.setIsInterior(false);
				
				Edge ijE = EdgeHelper.createEdge(nextOrigin, jN, depot);
				nextOrigin.setIsInterior(false);

				r.getEdges().set(r.getEdges().indexOf(e), ijE);
				r.addCosts(ijE);
				r.removeCosts(e);
				
				iNc.setDiEdge(ijE.getInverseEdge());
				iNc.setIdEdge(ijE);
			
			} else 
				if (jN instanceof SubCustomer) {
					SubCustomer jNs = (SubCustomer)jN;
					Customer jNc = jNs.getCustomer();
					
					r.getDeliveries().add(new Delivery(jNc, jNs.getProduct()));
					
					if(visitedNodes.contains(jNc)) 
						it.remove();
					else {
						Edge ijE = EdgeHelper.createEdge(nextOrigin, jNc, depot);
						jNc.setIsInterior(nextOrigin != depot);
						r.getEdges().set(r.getEdges().indexOf(e), ijE);
						r.addCosts(ijE);
						
						nextOrigin = jNc;
						
						Edge diEdge = EdgeHelper.createEdge(depot, jNc, depot);
						jNc.setDiEdge(diEdge);
						jNc.setIdEdge(diEdge.getInverseEdge());
						
						visitedNodes.add(jNc);
					}
					r.removeCosts(e);
				}
		}
		return r;
	}

	private SubSol constructSubSolution(List<Route> routes, SubMap sm) {
		int diffProducts = sm.getDepot().getDiffProducts();
		SubSol subSol = new SubSol(diffProducts);
		for (Route r : routes) {
			subSol.addCosts(r);
			subSol.getRoutes().add(r);
			for (int i = 0; i < diffProducts; i++)
				subSol.setProductDemand(i, subSol.getProductDemand(i) + r.getProductDemand(i));
		}
		subSol.assignSubMap(sm);
		return subSol;
	}

	private CompleteMap genRandAssignmentMap(List<Customer> nodes,
			List<Depot> depots, Savings savings) {
		int numNodes = nodes.size();

		CompleteMap bigMap = new CompleteMap();
		
		for (Depot depot : depots) 
			bigMap.addSubMap(new SubMap(depot, input.getMaxTime()) );

		class Mapper {
			boolean assigned[] ;
			int[] round ; // counter for depots
			int nAssignedNodes;
			Integer[][] savingsMatrix;
			int[][] randomArray;
			
			public Mapper(Savings savings) {
				int numNodes = input.getNumCustomers();
				int numDepots = input.getNumDepots();
				assigned = new boolean[numNodes];
				round = new int[numDepots];
				nAssignedNodes = 0;
					
				/*
				 * Create an array of lists of nodes, one list for each submap
				 * (identified by its depotId), each list sorted by savings (decreasing)
				 */
				savingsMatrix = new Integer[numDepots][numNodes];
	
				for (int i = 0; i < numDepots; i++) {
					for (int j = 0; j < numNodes; j++)
						savingsMatrix[i][j] = j;
					Arrays.sort(savingsMatrix[i], savings.getSavingsComparator(i));
				}
	
				/*
				 * Create array for random
				 */
	
				randomArray = new int[numDepots][numNodes];
	
				for (int i = 0; i < numDepots; i++) 
					randomArray[i] = RNG.getPositionsArrayGD(numNodes, test.getBeta());
			}
			/*
			 * Assigns the best node (influenced by bias-randomization) to the submap
			 */
			public boolean assignBestNode(SubMap sm, List<Customer> nodes) {
				int depotId = sm.getDepotId();
				
				int pos = randomArray[depotId][round[depotId]];
				int nodeId = savingsMatrix[depotId][pos];
				
				Customer currentNode = nodes.get(nodeId);
				
				if (!assigned[nodeId] && sm.canSupply(currentNode)) {
					sm.addNode(currentNode);
					assigned[nodeId] = true;
					nAssignedNodes++;
				}
				
				round[depotId]++;
				return round[depotId] == nodes.size();
			}
		}
		
		Mapper mapper = new Mapper(savings);
		
		List<SubMap> subMapList = new LinkedList<SubMap>(bigMap.getSubMaps());	
				
		while (mapper.nAssignedNodes < numNodes && !subMapList.isEmpty()) {
			Collections.sort(subMapList); // By % of free capacity in depot vehicles
						
			SubMap sm = subMapList.get(0);
			
			boolean isLastNodeOfDepot = mapper.assignBestNode(sm, nodes);
			
			if (isLastNodeOfDepot) subMapList.remove(sm);
		}
		
		if (mapper.nAssignedNodes < numNodes) return null;
		
		return bigMap;
	}

}
