package main;
/***********************************************************************************
 * Project ILS-MDVRP - TestsRunner.java 
 * This class reads the tests to be executed from TestPlanner class,
 * executes them and prints outputs to file.
 * Date of last revision (YYMMDD): 120222
 * (C) Angel A. Juan - http://ajuanp.wordpress.com
 **********************************************************************************/

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Scanner;

import javax.swing.JFrame;

import models.Depot;
import models.Customer;
import models.Test;
import models.VehType;
import models.Vehicle;
import models.Warehouse;
import plot.InstancePlot;
import RCWS.RouteCache;
import algorithm.Algorithm;
import data.Input;
import data.Solution;

public class TestsRunner {

	private static String _pathIn = "inputs"; // input files path (relative to project)
	private static String _pathOut = "results";
	private static String _testsFileName = "tests2run.txt";
	private static String suffix = ".txt";

	public static void main(String[] args) throws IOException {
		List<Test> tests = testLoad(_pathIn + "/" + _testsFileName);
		
		for (Test test : tests) {
			int seed = test.getSeed();
			String timeStamp = new SimpleDateFormat("yyyy.MM.dd.HH.mm.ss").format(new Date());
			String testValues = test.getInstanceName() + "-" + test.getBeta() + "-" 
					+ test.getSavingsParameter() + "-" + test.getTime() + "-" + test.getMaxTime();
			PrintStream fileStream = new PrintStream(
					new File(_pathOut + "/T_" + timeStamp + "_" + testValues + suffix));
			
			for (int i = 0; i<test.getnTimes(); i++) {
				
				test.setSeed(seed+i);
				
				Input in = loadInstance(_pathIn + "/" + test.getInstanceName()
						+ suffix, test.getMaxTime());
	
				new RouteCache();
				
				System.out.println("Start solving: " + test.getInstanceName() + 
						" Mapping beta: " + test.getBeta() + " CWS beta: " + 
						test.getSavingsParameter() );
				
				Algorithm alg = new Algorithm(test, in);
				Solution sol = alg.solve();
				
				if (sol==null) {
					System.out.println("No solution found");
				} else {
					System.out.println("\n\n" + sol);
					fileStream.printf(Locale.GERMANY,"%.2f;%.2f;%.2f\n",in.getMaxTime(),sol.getCost(),sol.getTimeLength());
					fileStream.flush();
					//miniplot of route.......
					InstancePlot p = new InstancePlot(sol, in);
					p.plot();
				}
			}
			fileStream.close();
		}
		
	}

	public static Input loadInstance(String fileName, float testMaxTime) {
		Input input = null;
		BufferedReader br;
		try {
			br = new BufferedReader(new FileReader(fileName));
			Scanner in = new Scanner(br);
			in.useLocale(Locale.US);
			
			String s = "#";
			while (s.startsWith("#")) s = in.nextLine();
			
			// Main instance parameters
			Scanner in2 = new Scanner(s);
			in2.useLocale(Locale.US);

			int customers = in2.nextInt();
			int depots = in2.nextInt();
			int warehouses = in2.nextInt();
			int diffProducts = in2.nextInt();
			int diffVehicles = in2.nextInt();
			int distType = in2.nextInt();
			float maxTime = in2.nextFloat();
			if (testMaxTime > 0f) maxTime = testMaxTime;
			float speed = in2.nextFloat();
			float flow = in2.nextFloat();
			float opSetupTime = in2.nextFloat();
			float intraCityTime = in2.nextFloat();
			
			int nodes = customers+depots+warehouses;
			
			List<Customer> customersList = new LinkedList<Customer>();
			List<Depot> depotsList = new LinkedList<Depot>();
			List<Warehouse> warehousesList = new LinkedList<Warehouse>();
			List<VehType> vehTypeList = new LinkedList<VehType>();
			
			// Vehicle Types for this instance
			for (int i = 0; i < diffVehicles; i++) {
				int id = in.nextInt();
				double fixCost = in.nextDouble();
				double varCost = in.nextDouble();
				int whCapable = in.nextInt();
				String l = in.nextLine().trim();
				String[] items = l.split(" ");
				int[] caps = new int[items.length];
				for (int j = 0; j < items.length; j++)
					caps[j] = Integer.parseInt(items[j]);
				vehTypeList.add(new VehType(id, caps, fixCost, varCost, whCapable == 1));
			}
			
			// Customers
			for (int i = 0; i < customers; i++) {
				int id = in.nextInt();
				String locName = in.next();
				int x = in.nextInt();
				int y = in.nextInt();
				// Product demands
				int[] productsDemand = new int[diffProducts];
				for(int j = 0; j < diffProducts; j++) {
					productsDemand[j] = in.nextInt();
				}
				// Accepted Vehicle Types
				String l = in.nextLine().trim();
				String[] items = l.split(" ");
				boolean[] acceptedVehTypes = new boolean[diffVehicles];
				for(int j = 0; j < items.length; j++) { 
					int vt = Integer.parseInt(items[j]);  
					acceptedVehTypes[vt] = true; 		
				}
				
				Customer node = new Customer(id, x, y, productsDemand, acceptedVehTypes, locName);
				customersList.add(node);
			}
			
			// Depots
			for (int i = 0; i < depots; i++) {
				int id = in.nextInt();
				String locName = in.next();
				int x = in.nextInt();
				int y = in.nextInt();
				int[] productsCapacity = new int[diffProducts];
				for(int j = 0; j < diffProducts; j++) {
					productsCapacity[j] = in.nextInt();
				}
				int[] productsReserve = new int[diffProducts];
				for(int j = 0; j < diffProducts; j++) {
					productsReserve[j] = in.nextInt();
				}
				Depot depot = new Depot(id, i, x, y, productsCapacity, productsReserve, locName);
				
				String l = in.nextLine().trim();
				String[] items = l.split(" ");
				
				for(int j = 0; j < items.length / 2; j++) {  
					
					int qty = Integer.parseInt(items[2*j]);  		// qty of vehicles of this type

					int vehTypeNo = Integer.parseInt(items[2*j+1]); // Id of VehicleType
					
					if (qty>0) {
						Vehicle veh = new Vehicle(vehTypeList.get(vehTypeNo), qty);
						veh.assignTo(depot);
					}
				}
				depotsList.add(depot);
				
			}
			
			// Warehouses
			for (int i = 0; i < warehouses; i++) {
				int id = in.nextInt();
				String locName = in.next();
				int x = in.nextInt();
				int y = in.nextInt();
				Warehouse w = new Warehouse(id, x, y, locName);
				warehousesList.add(w);
			}
			
			float[][] distances = new float[nodes][nodes];
			float[][] times = new float[nodes][nodes];
			if (distType != 0) {
			//Distance matrix
			
				for (int i = 1; i < nodes; i++)
					for (int j = 0; j < i; j++) {
						float d = in.nextFloat();
						distances[i][j] = d;
						distances[j][i] = d;
					}
			// Time matrix
				for (int i = 1; i < nodes; i++)
					for (int j = 0; j < i; j++) {
						float d = in.nextFloat();
						times[i][j] = d;
						times[j][i] = d;
					}
			}
			input = new Input(customersList, depotsList, warehousesList, diffProducts, 
					distances, times, maxTime, distType, speed, flow, opSetupTime,
					intraCityTime);
			in2.close();
			in.close();
		} catch (FileNotFoundException e) {
		
			e.printStackTrace();
		}

		return input;
	}

	public static List<Test> testLoad(String filePath) throws FileNotFoundException {
		List<Test> tests = new LinkedList<Test>();
		
		BufferedReader br = new BufferedReader(new FileReader(filePath));
		Scanner in = new Scanner(br);
		in.useLocale(Locale.US);
		while (in.hasNext()) {
			String s = in.nextLine();
			if (!s.startsWith("#")) {
				Scanner in2 = new Scanner(s);
				in2.useLocale(Locale.US);
				String name = in2.next();
				double betaMap = in2.nextDouble();
				double betaCWS = in2.nextDouble();
				int seed = in2.nextInt();
				int time = in2.nextInt();
				float maxTime = in2.nextFloat();
				int nTimes = in2.nextInt();
				Test test = new Test(name, betaMap, betaCWS, seed, time, maxTime, nTimes);
				tests.add(test);
				in2.close();
			}
		}
		in.close();
		

		return tests;
	}
}
