package map;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

import data.Input;
import algorithm.ArrayTools;
import RCWS.Edge;
import RCWS.Route;
import RCWS.RouteGroup;
import RCWS.WhRoute;
import main.RNG;
import models.Availability;
import models.Delivery;
import models.Depot;
import models.Customer;
import models.Node;
import models.VehType;
import models.Vehicle;
import models.Warehouse;

public class SubMap implements Comparable<SubMap> {
	
	static long nInstance = 0;
	private long id;
	private Depot depot;
	private List<Node> nodesInSubMap;
	
	private int[] demand = null;
	private final int sumOfAllVehCapacities;
	private Set<Delivery> deliveryNeeds;
	private Map<VehType, List<RouteGroup>> routeClassif;
	private List<Availability> avails ;
	
	float maxTime;
	
	
	public SubMap(Depot depot, float maxTime) {
		this.id = nInstance++;
		this.depot = depot;
		demand = new int[depot.getDiffProducts()];
		nodesInSubMap = new LinkedList<Node>();
		routeClassif = new HashMap<VehType,List<RouteGroup>>();
		int tmp = 0;
		for(Vehicle v : depot.getVehicles()) {
			tmp += v.getTotalCapacity() * v.getNumVehicles();
			routeClassif.put(v.getVehType(), new ArrayList<RouteGroup>());
		}
		sumOfAllVehCapacities = tmp;
		deliveryNeeds = new HashSet<Delivery>();
		this.avails = new LinkedList<Availability>();
		if (ArrayTools.gtZero(depot.getProductsCapacity()))
			avails.add(new Availability(null, 0, depot.getProductsCapacity()));
		this.maxTime = maxTime;
	}
	
	public long getId() {
		return id;
	}

	public int getSumOfAllVehCapacities() {
		return sumOfAllVehCapacities;
	}
	
	public Set<Delivery> getDeliveryNeeds() {
		return deliveryNeeds;
	}

	public int[] getDemand() {
		return demand;
	}
	
	public List<Availability> getAvails() {
		return avails;
	}

	public int sumOfAllDemands() {
		int acc = 0;
		for(int i = 0; i < depot.getDiffProducts(); i++) {
			acc += demand[i];
		}
		return acc;
	}
	
	public void addNode(Customer node) {
		nodesInSubMap.add(node);
		for(int i = 0; i < depot.getDiffProducts(); i++) {
			demand[i] += node.getDemand(i);
			if (node.getDemand(i) > 0)
				deliveryNeeds.add(new Delivery(node,i));
		}
	}
	
	public int getRemainingVehCapacity() {
		return sumOfAllVehCapacities - sumOfAllDemands(); 
	}
	
	public int getDepotId() { 
		return depot.getDepotId();
	}
	
	public Depot getDepot() {
		return depot;
		
	}
	public Map<VehType, List<RouteGroup>> getRouteClassif() {
		return routeClassif;
	}

	@Override
	public int compareTo(SubMap o) {

		float othersRatio = (float)o.getRemainingVehCapacity()/o.getSumOfAllVehCapacities(); 
		float myRatio = (float)this.getRemainingVehCapacity()/this.getSumOfAllVehCapacities();
		
		if (othersRatio > myRatio ) return 1;
		if (othersRatio < myRatio ) return -1;
		return 0;
	}
	
	public List<Node> getAssignedNodes() {
		return nodesInSubMap;
	}

	public boolean canSupply(Customer c) {		
		// Exists one vehicle in sm depot that can suply c
		for (Vehicle v : this.getDepot().getVehicles()) 
			if (c.acceptsVehType(v.getVehType()))
				return true;	
		
		return false;
	}
	/**
	 * checks if last depot avalaibility can supply a given demand  
	 * @param demand
	 * @return true if it's enough, false otherwise
	 */
	public boolean enoughInventory(int[] demand) {
		return !ArrayTools.gtZero(ArrayTools.diff(demand, getRemainingDepotInventory()));
		// demand <= sm.getRemainingDepotInventory()
	}

	/**
	 * Increments affected availabilities when a wh-route arrives
	 * @param whR
	 */
	public void addAvail(WhRoute whR) {
		// qtys aported by whR
		int[] qtys = ArrayTools.diff(whR.getPurchase(), whR.getDemand());
		if (ArrayTools.gtZero(qtys)) {
			// Look for position where to insert, preserving time order
			int i = avails.size();
			float arrivalTime = whR.getStartTime() + whR.getTimeLength();
			boolean found = false;
			while (i>0 && !found)
				if (avails.get(i-1).getT() > arrivalTime) 
					i--;
				else
					found = true;
			// Insert new Availability with same qtys as previous one
			int newAvails[] ;
			
			if (i==0)
				newAvails = new int[depot.getDiffProducts()];
			else
				newAvails = avails.get(i-1).getQtys().clone();
			
			avails.add(i, new Availability(whR, arrivalTime, newAvails ));
			// Increment new and subsequent availabilities
			for (int j = i; j < avails.size(); j++ ) 
				ArrayTools.inc(avails.get(j).getQtys(), qtys);
		}
	}
	/**
	 * Decrements all affected availabilities when a Route departs
	 * @param r The added Route
	 * @param goodAv The first Availability that is enough for r
	 */
	public void substractAvail(Route r, Availability goodAv) {
		int[] qtys = r.getDemand();
		// Substract demand to last Availability --> new remaining
		int[] newQtys = ArrayTools.diff(getRemainingDepotInventory(), qtys);
		for (int prod = 0; prod < depot.getDiffProducts(); prod++) {
			int newProdQty = newQtys[prod];
			ListIterator<Availability> it = avails.listIterator(avails.size());
			// For each Availability
			boolean beforeGoodAv = false;
			while (it.hasPrevious()){
				Availability a = it.previous();
				
				if (beforeGoodAv) {
					// The availability is before goodAv --> decrement only to reach new qty 
					if (a.getQty(prod) > newProdQty)
						a.setQty(prod, newProdQty);
				} else {
					// The availability is goodAv or after it --> decrement demand
					newProdQty = a.getQty(prod) - qtys[prod];
					a.setQty(prod, newProdQty);
				}
				if (a == goodAv) beforeGoodAv = true;
			}
		}
		
		// Remove null availabilities
		Iterator<Availability> it = avails.iterator();
		boolean removed = true;
		while (it.hasNext() && removed) {
			Availability a = it.next();
			removed = !ArrayTools.gtZero(a.getQtys());
			if (removed) 
				it.remove();
		}
	}

	/**
	 * Finds first Availability enough for a demand
	 * @param demands
	 * @return Found Availability or null if not found
	 */
	public Availability getAvailability(int[] demands) {
		boolean found = false;
		Availability a = null;
		Iterator<Availability> it = avails.iterator();
		while(it.hasNext() && !found) {
			a = it.next();
			found = !ArrayTools.gtZero(ArrayTools.diff(demands, a.getQtys()));
			      // qtys >= demands
		}
		
		if(!found) 
			return null;
		return a;
	}
		
	public int[] getRemainingDepotInventory() {
		if (avails.isEmpty()) return new int[depot.getDiffProducts()];
		return avails.get(avails.size()-1).getQtys();
	}

	/**
	 * Classifies a route generated by CWS according to available resources.
	 * 
	 * @param r The Route
	 * @param veh Vehicle
	 * @return Route could be classified
	 */
	public boolean putRoute(Route r, Vehicle veh) {
		
		VehType vt = veh.getVehType();
		List<RouteGroup> routeGroups = getRouteClassif().get(vt);
		boolean leftVehicles = routeGroups.size() < veh.getNumVehicles();
		boolean ok = false;

		if (r instanceof WhRoute) {
			// Route first visits warehouse ==> increases inventory
			WhRoute whR = (WhRoute)r;
			
			if (leftVehicles && r.getTimeLength() < maxTime) {
	
				// Create new group of routes for this vehicle. No dependencies.
				RouteGroup rg = new RouteGroup();
				whR.setStartTime(0f);
				rg.addRoute(whR);
				routeGroups.add(rg);
				ok = true;
			} else {
				// No more vehicles of this type, so let's try to add the route to an existing group
				float minTime = maxTime + 1;
				// Find the group that allows an earlier finish
				RouteGroup minRg = null;
				float newTimeLength = 0;
				for (RouteGroup rg : routeGroups) {
					newTimeLength = rg.getEndTime() + r.getTimeLength();
					if (newTimeLength < minTime) {
						minTime = newTimeLength;
						minRg = rg;
					}
				}
				// Add the Route to the RouteGroup
				if (minRg != null && newTimeLength < maxTime) {
					whR.setStartTime(minRg.getEndTime());
					minRg.addRoute(whR);
					ok = true;
				}
			}
		} else {
			// route doesn't visit wh ==> decreases inventory
			// Check if route demand fits in any availability	
			Availability a = getAvailability(r.getDemand());
			
			if (a != null) 				
				if (a.getT() + r.getTimeLength() < maxTime) {
				// 0.5 probability to start a new group
					int rnd = RNG.getNextIntUD(10);
					
					if (!leftVehicles || rnd < 5) 					
						ok = putInExistingGroup(routeGroups, a, r, maxTime);
					
					if (!ok && leftVehicles) {
						putInNewGroup (routeGroups, a, r);
						ok = true;
					}
				}
		}
		return ok;
	}
	/**
	 * puts a route in an existing group
	 * @param routeGroups
	 * @param a
	 * @param route
	 * @param maxTime
	 * @return
	 */
	private boolean putInExistingGroup(List<RouteGroup> routeGroups, Availability a, Route route, float maxTime) {
		Collections.shuffle(routeGroups, RNG.getrng());
		boolean found = false;
		for(int i = 0; i < routeGroups.size() && !found; i++) {
			RouteGroup rg = routeGroups.get(i);

			// Check if route fits in time
			
			float startTime = Math.max(rg.getEndTime(), a.getT());
			found = startTime + route.getTimeLength() <= maxTime;
			if (found) {

				// Add the route into the group
				route.setStartTime(startTime);
				rg.addRoute(route);
				//Decrease depot's availabilities
				substractAvail(route, a);	
				return true;
			}	
		}		
		return false;
	}


	/**
	 * Opens a new Route group
	 * @param routeGroups
	 * @param a
	 * @param r
	 * @return
	 */
	private void putInNewGroup(List<RouteGroup> routeGroups, Availability a, Route r) {
		
		RouteGroup rg = new RouteGroup();
		r.setStartTime(a.getT());
		rg.addRoute(r);
		routeGroups.add(rg);
		
		// Decrease depot's availabilities
		substractAvail(r, a);
	}
	
	/**
	 * Calculates earliest start time for a Route
	 * @param veh The Route's Vehicle
	 * @param r The Route
	 * @param demand The Route's demand
	 * @return The start time
	 */
	public float calcStartTime(Vehicle veh, Route r, int[] demand) {
		float minTime = 0f;
		
		List<RouteGroup> routeGroups = getRouteClassif().get(veh.getVehType());
		
		if(routeGroups.size() == veh.getNumVehicles()) {	
			// Cannot start new group: find the shortest group in time
			minTime = maxTime + 1f;
			for (RouteGroup rg : routeGroups) 
				if (rg.getEndTime() < minTime) 
					minTime = rg.getEndTime();
		} else 
			if (r instanceof WhRoute) 
				// If WhRoute can start its own group, startTime = 0
				return 0f;
		
		// Calc time when demand is available in Depot
		float availTime = 0f;
		if (!(r instanceof WhRoute)) {
			Availability a = getAvailability(demand);
			if (a == null) 
				availTime = maxTime + 1f;
			else
				availTime = a.getT();
		}
		return Math.max(availTime, minTime);
	}
	
	public float calcStartTime(Vehicle veh, WhRoute whRoute) {
		return calcStartTime(veh, whRoute, whRoute.getDemand());
	}

	/**
	 * Calculates a warehouse to supply a set of nodes. Sorts all
	 * warehouses according the sum of distances to the nodes and chooses, in a 
	 * bias randomized way with beta=0.5, the "best"
	 * @param nodes
	 * @param warehouses
	 * @return the "best" warehouse
	 */
	public Warehouse getAwh(List<Node> nodes, List<Warehouse> warehouses) {
		class WhDist implements Comparable<WhDist> {
			Warehouse w; Double d;
			public WhDist(Warehouse wh, Double sumDist) { w = wh; d = sumDist;}
			public int compareTo(WhDist other) {
				if (this.d > other.d) return 1;
				if (this.d < other.d) return -1;
				return 0;
			}
		}
		Warehouse result = null;
		switch (warehouses.size()) {
		case 0: break;
		case 1: return warehouses.get(0);
		default:
			List<WhDist> list = new ArrayList<WhDist>();
			for (Warehouse wh : warehouses) {
				Double sumDist = 0.;
				for (Node n : nodes) 
					sumDist += MapTools.distanceMatrix[wh.getId()][n.getId()];
				list.add(new WhDist(wh, sumDist));
			}
			Collections.sort(list);
			int index = RNG.getNextIntGD(list.size(), 0.5);
			result = list.get(index).w;
		}
		return result;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("SubMap [id=").append(id)
				.append(", depot=").append(this.depot)
				.append(",\n\t\t\t\tnodes=[")
				.append("\n\t\t\t\t\t").append(MapTools.str(nodesInSubMap, 5, 100))
				.append("\n\t\t\t\t]")
				.append(",\n\t\t\t\tdemand=").append(Arrays.toString(demand))
				.append(",\n\t\t\t\tdeliveryNeeds=[")
				.append("\n\t\t\t\t\t").append(MapTools.str(deliveryNeeds, 5, 100))
				.append("\n\t\t\t\t]")
				.append(",\n\t\t\t\trouteClassif=").append(str(routeClassif))
				.append(",\n\t\t\t\tavails=[");
				if (!avails.isEmpty())
					builder.append(MapTools.str(avails, 0, 0))
					.append("\n\t\t\t\t");
				builder.append("]")
				.append("\n\t\t\t]");
		return builder.toString();
	}

	private String str(Map<VehType, List<RouteGroup>> rc) {
		StringBuilder builder = new StringBuilder();
		builder.append ("{");
		Iterator<VehType> it = rc.keySet().iterator();
		if (it.hasNext()) {
			VehType vt = it.next();
			if (!rc.get(vt).isEmpty()) 
				builder.append("\n\t\t\t\t\t")
						.append(vt).append("=[").append(MapTools.str(rc.get(vt),0, 0))
						.append("\n\t\t\t\t\t]");
			while (it.hasNext()) {
				vt = it.next();
				if (!rc.get(vt).isEmpty()) 
					builder.append(",\n\t\t\t\t\t")
							.append(vt).append("=[").append(MapTools.str(rc.get(vt),0, 0))
							.append("\n\t\t\t\t\t]");
			}
		}
		builder.append("\n\t\t\t\t}");
		return builder.toString();
	}
}
